@@ -155,27 +155,35 @@ impl ApplicationHandler<EventPayload> for Application<'_> {
             RioEventType::Rio(RioEvent::RenderRoute(route_id)) => {
                 if self.config.renderer.strategy.is_event_based() {
                     if let Some(route) = self.router.routes.get_mut(&window_id) {
+                        // Skip rendering for unfocused windows if configured
                         if self.config.renderer.disable_unfocused_render
                             && !route.window.is_focused
                         {
                             return;
                         }
 
-                        // TODO: Optionally Allow check if is the current route
-                        // before schedule render. Using the following code below:
-                        // if route_id == route.window.screen.ctx().current_route() {
-                        let timer_id = TimerId::new(Topic::RenderRoute, route_id);
-                        let event = EventPayload::new(
-                            RioEventType::Rio(RioEvent::Render),
-                            window_id,
-                        );
+                        // Check if this is the current route
+                        if route_id == route.window.screen.ctx().current_route() {
+                            // Check if we need to throttle based on timing
+                            if let Some(wait_duration) = route.window.wait_until() {
+                                // We need to wait before rendering again
+                                let timer_id = TimerId::new(Topic::RenderRoute, route_id);
+                                let event = EventPayload::new(
+                                    RioEventType::Rio(RioEvent::Render),
+                                    window_id,
+                                );
 
-                        if !self.scheduler.scheduled(timer_id) {
-                            if let Some(limit) = route.window.wait_until() {
-                                route.window.start_render_timestamp();
-                                self.scheduler.schedule(event, limit, false, timer_id);
+                                // Only schedule if not already scheduled
+                                if !self.scheduler.scheduled(timer_id) {
+                                    self.scheduler.schedule(
+                                        event,
+                                        wait_duration,
+                                        false,
+                                        timer_id,
+                                    );
+                                }
                             } else {
-                                route.window.start_render_timestamp();
+                                // We can render immediately
                                 route.request_redraw();
                             }
                         }
@@ -198,28 +206,6 @@ impl ApplicationHandler<EventPayload> for Application<'_> {
                     );
                 }
             }
-            RioEventType::Rio(RioEvent::UpdateGraphicLibrary) => {
-                if let Some(route) = self.router.routes.get_mut(&window_id) {
-                    let mut terminal = route
-                        .window
-                        .screen
-                        .context_manager
-                        .current_mut()
-                        .terminal
-                        .lock();
-                    let graphics = terminal.graphics_take_queues();
-                    drop(terminal);
-                    if let Some(graphic_queues) = graphics {
-                        for graphic_data in graphic_queues.pending {
-                            route.window.screen.sugarloaf.graphics.insert(graphic_data);
-                        }
-
-                        for graphic_data in graphic_queues.remove_queue {
-                            route.window.screen.sugarloaf.graphics.remove(&graphic_data);
-                        }
-                    }
-                }
-            }
             RioEventType::Rio(RioEvent::ReportToAssistant(error)) => {
                 if let Some(route) = self.router.routes.get_mut(&window_id) {
                     route.report_error(&error);
@@ -1123,8 +1109,11 @@ impl ApplicationHandler<EventPayload> for Application<'_> {
             }
 
             WindowEvent::RedrawRequested => {
+                // let start = std::time::Instant::now();
                 route.window.winit_window.pre_present_notify();
 
+                route.begin_render();
+
                 match route.path {
                     RoutePath::Assistant => {
                         route.window.screen.render_assistant(&route.assistant);
@@ -1143,14 +1132,13 @@ impl ApplicationHandler<EventPayload> for Application<'_> {
                         );
                     }
                 }
+
+                // let duration = start.elapsed();
                 // println!("Time elapsed in render() is: {:?}", duration);
                 // }
 
-                if self.config.renderer.strategy.is_game()
-                    && !self.config.renderer.disable_unfocused_render
-                    && route.window.is_focused
-                {
-                    route.request_frame(&mut self.scheduler);
+                if self.config.renderer.strategy.is_game() {
+                    route.request_redraw();
                 }
 
                 event_loop.set_control_flow(ControlFlow::Wait);
