@@ -3,22 +3,25 @@ mod search;
 pub mod utils;
 
 use crate::ansi::CursorShape;
-use crate::context::renderable::{Cursor, RenderableContent, RenderableContentStrategy};
+use crate::context::renderable::{Cursor, RenderableContent};
 use crate::context::ContextManager;
 use crate::crosswords::grid::row::Row;
 use crate::crosswords::pos::{Column, Line, Pos};
 use crate::crosswords::square::{Flags, Square};
 use crate::screen::hint::HintMatches;
 use navigation::ScreenNavigation;
+use rio_backend::ansi::graphics::UpdateQueues;
+use rio_backend::config::colors::term::TermColors;
 use rio_backend::config::colors::{
     term::{List, DIM_FACTOR},
     AnsiColor, ColorArray, Colors, NamedColor,
 };
 use rio_backend::config::Config;
+use rio_backend::crosswords::TermDamage;
 use rio_backend::event::EventProxy;
 use rio_backend::sugarloaf::{
-    Content, FragmentStyle, FragmentStyleDecoration, Graphic, Stretch, Style,
-    SugarCursor, Sugarloaf, UnderlineInfo, UnderlineShape, Weight,
+    drawable_character, Content, FragmentStyle, FragmentStyleDecoration, Graphic,
+    Stretch, Style, SugarCursor, Sugarloaf, UnderlineInfo, UnderlineShape, Weight,
 };
 use std::collections::HashMap;
 use std::ops::RangeInclusive;
@@ -35,10 +38,12 @@ pub struct Search {
 pub struct Renderer {
     is_vi_mode_enabled: bool,
     draw_bold_text_with_light_colors: bool,
+    use_drawable_chars: bool,
     pub named_colors: Colors,
     pub colors: List,
     pub navigation: ScreenNavigation,
     unfocused_split_opacity: f32,
+    last_active: usize,
     pub config_has_blinking_enabled: bool,
     pub config_blinking_interval: u64,
     ignore_selection_fg_color: bool,
@@ -87,6 +92,8 @@ impl Renderer {
 
         Renderer {
             unfocused_split_opacity: config.navigation.unfocused_split_opacity,
+            last_active: 0,
+            use_drawable_chars: config.fonts.use_drawable_chars,
             draw_bold_text_with_light_colors: config.draw_bold_text_with_light_colors,
             macos_use_unified_titlebar: config.window.macos_use_unified_titlebar,
             config_blinking_interval: config.cursor.blinking_interval.clamp(350, 1200),
@@ -117,13 +124,12 @@ impl Renderer {
     fn create_style(
         &mut self,
         square: &Square,
-        renderable_content: &RenderableContent,
+        term_colors: &TermColors,
     ) -> (FragmentStyle, char) {
         let flags = square.flags;
 
-        let mut foreground_color =
-            self.compute_color(&square.fg, flags, renderable_content);
-        let mut background_color = self.compute_bg_color(square, renderable_content);
+        let mut foreground_color = self.compute_color(&square.fg, flags, term_colors);
+        let mut background_color = self.compute_bg_color(square, term_colors);
 
         let content = if square.c == '\t' || flags.contains(Flags::HIDDEN) {
             ' '
@@ -156,8 +162,7 @@ impl Renderer {
             Some(background_color)
         };
 
-        let (decoration, decoration_color) =
-            self.compute_decoration(square, renderable_content);
+        let (decoration, decoration_color) = self.compute_decoration(square, term_colors);
 
         (
             FragmentStyle {
@@ -176,7 +181,7 @@ impl Renderer {
     fn compute_decoration(
         &self,
         square: &Square,
-        renderable_content: &RenderableContent,
+        term_colors: &TermColors,
     ) -> (Option<FragmentStyleDecoration>, Option<[f32; 4]>) {
         let mut decoration = None;
         let mut decoration_color = None;
@@ -223,7 +228,7 @@ impl Renderer {
         if decoration.is_some() {
             if let Some(color) = square.underline_color() {
                 decoration_color =
-                    Some(self.compute_color(&color, square.flags, renderable_content));
+                    Some(self.compute_color(&color, square.flags, term_colors));
             }
         };
 
@@ -242,8 +247,10 @@ impl Renderer {
         renderable_content: &RenderableContent,
         search_hints: &mut Option<HintMatches>,
         focused_match: &Option<RangeInclusive<Pos>>,
+        term_colors: &TermColors,
         is_active: bool,
     ) {
+        // let start = std::time::Instant::now();
         let cursor = &renderable_content.cursor;
         let hyperlink_range = renderable_content.hyperlink_range;
         let selection_range = renderable_content.selection_range;
@@ -259,13 +266,12 @@ impl Renderer {
                 continue;
             }
 
-            let (mut style, square_content) = if has_cursor
-                && column == cursor.state.pos.col
-            {
-                self.create_cursor_style(square, cursor, is_active, renderable_content)
-            } else {
-                self.create_style(square, renderable_content)
-            };
+            let (mut style, square_content) =
+                if has_cursor && column == cursor.state.pos.col {
+                    self.create_cursor_style(square, cursor, is_active, term_colors)
+                } else {
+                    self.create_style(square, term_colors)
+                };
 
             if hyperlink_range.is_some()
                 && square.hyperlink().is_some()
@@ -286,7 +292,7 @@ impl Renderer {
                     .contains(Pos::new(line, Column(column)))
             {
                 style.color = if self.ignore_selection_fg_color {
-                    self.compute_color(&square.fg, square.flags, renderable_content)
+                    self.compute_color(&square.fg, square.flags, term_colors)
                 } else {
                     self.named_colors.selection_foreground
                 };
@@ -338,42 +344,95 @@ impl Renderer {
                 style.background_color = None;
             }
 
-            if let Some((font_id, width)) =
-                self.font_cache.get(&(square_content, style.font_attrs))
-            {
-                style.font_id = *font_id;
-                style.width = *width;
-            } else {
-                let mut width = square.c.width().unwrap_or(1) as f32;
-                let mut font_ctx = self.font_context.inner.lock();
-
-                // There is no simple way to define what's emoji
-                // could have to refer to the Unicode tables. However it could
-                // be leading to misleading results. For example if we used
-                // unicode and internationalization functionalities like
-                // https://github.com/open-i18n/rust-unic/, then characters
-                // like "◼" would be valid emojis. For a terminal context,
-                // the character "◼" is not an emoji and should be treated as
-                // single width. So, we completely rely on what font is
-                // being used and then set width 2 for it.
-                if let Some((font_id, is_emoji)) =
-                    font_ctx.find_best_font_match(square_content, &style)
+            // In case we have a drawable character then we will:
+            // 1. Ignore font cache and find width since it's known already.
+            // 2. Wrap up the content and send to sugarloaf.
+            //
+            // TODO: In the future it should use same logic to render everything
+            // at once.
+            if self.use_drawable_chars {
+                if let Some(character) = drawable_character(square_content) {
+                    style.drawable_char = Some(character);
+                    if !content.is_empty() {
+                        if let Some(line) = line_opt {
+                            builder.add_text_on_line(line, &content, last_style);
+                        } else {
+                            builder.add_text(&content, last_style);
+                        }
+                        content.clear();
+                    }
+
+                    last_style = style;
+
+                    // Ignore font shaping
+                    content.push(' ');
+                }
+            }
+
+            let has_drawable_char = style.drawable_char.is_some();
+            if !has_drawable_char {
+                if let Some((font_id, width)) =
+                    self.font_cache.get(&(square_content, style.font_attrs))
                 {
-                    style.font_id = font_id;
-                    if is_emoji {
-                        width = 2.0;
+                    style.font_id = *font_id;
+                    style.width = *width;
+                } else {
+                    let mut width = square.c.width().unwrap_or(1) as f32;
+                    let mut font_ctx = self.font_context.inner.lock();
+
+                    // There is no simple way to define what's emoji
+                    // could have to refer to the Unicode tables. However it could
+                    // be leading to misleading results. For example if we used
+                    // unicode and internationalization functionalities like
+                    // https://github.com/open-i18n/rust-unic/, then characters
+                    // like "◼" would be valid emojis. For a terminal context,
+                    // the character "◼" is not an emoji and should be treated as
+                    // single width. So, we completely rely on what font is
+                    // being used and then set width 2 for it.
+                    if let Some((font_id, is_emoji)) =
+                        font_ctx.find_best_font_match(square_content, &style)
+                    {
+                        style.font_id = font_id;
+                        if is_emoji {
+                            width = 2.0;
+                        }
                     }
+                    style.width = width;
+
+                    self.font_cache.insert(
+                        (square_content, style.font_attrs),
+                        (style.font_id, style.width),
+                    );
+                };
+
+                if square_content == ' ' {
+                    if !last_char_was_space {
+                        if !content.is_empty() {
+                            if let Some(line) = line_opt {
+                                builder.add_text_on_line(line, &content, last_style);
+                            } else {
+                                builder.add_text(&content, last_style);
+                            }
+                            content.clear();
+                        }
+
+                        last_char_was_space = true;
+                        last_style = style;
+                    }
+                } else {
+                    if last_char_was_space && !content.is_empty() {
+                        if let Some(line) = line_opt {
+                            builder.add_text_on_line(line, &content, last_style);
+                        } else {
+                            builder.add_text(&content, last_style);
+                        }
+                        content.clear();
+                    }
+
+                    last_char_was_space = false;
                 }
-                style.width = width;
-
-                self.font_cache.insert(
-                    (square_content, style.font_attrs),
-                    (style.font_id, style.width),
-                );
-            };
 
-            if square_content == ' ' {
-                if !last_char_was_space {
+                if last_style != style {
                     if !content.is_empty() {
                         if let Some(line) = line_opt {
                             builder.add_text_on_line(line, &content, last_style);
@@ -383,37 +442,12 @@ impl Renderer {
                         content.clear();
                     }
 
-                    last_char_was_space = true;
                     last_style = style;
                 }
-            } else {
-                if last_char_was_space && !content.is_empty() {
-                    if let Some(line) = line_opt {
-                        builder.add_text_on_line(line, &content, last_style);
-                    } else {
-                        builder.add_text(&content, last_style);
-                    }
-                    content.clear();
-                }
 
-                last_char_was_space = false;
+                content.push(square_content);
             }
 
-            if last_style != style {
-                if !content.is_empty() {
-                    if let Some(line) = line_opt {
-                        builder.add_text_on_line(line, &content, last_style);
-                    } else {
-                        builder.add_text(&content, last_style);
-                    }
-                    content.clear();
-                }
-
-                last_style = style;
-            }
-
-            content.push(square_content);
-
             // Render last column and break row
             if column == (columns - 1) {
                 if !content.is_empty() {
@@ -433,6 +467,12 @@ impl Renderer {
         } else {
             builder.new_line();
         }
+
+        // let duration = start.elapsed();
+        // println!(
+        //     "Time elapsed in --renderer.update.create_line() is: {:?}",
+        //     duration
+        // );
     }
 
     #[inline]
@@ -440,7 +480,7 @@ impl Renderer {
         &self,
         color: &AnsiColor,
         flags: Flags,
-        renderable_content: &RenderableContent,
+        term_colors: &TermColors,
     ) -> ColorArray {
         match color {
             AnsiColor::Named(ansi) => {
@@ -453,18 +493,18 @@ impl Renderer {
                         if ansi == &NamedColor::Foreground
                             && self.named_colors.light_foreground.is_none() =>
                     {
-                        self.color(NamedColor::DimForeground as usize, renderable_content)
+                        self.color(NamedColor::DimForeground as usize, term_colors)
                     }
                     // Draw bold text in bright colors *and* contains bold flag.
                     (true, Flags::BOLD) => {
-                        self.color(ansi.to_light() as usize, renderable_content)
+                        self.color(ansi.to_light() as usize, term_colors)
                     }
                     // Cell is marked as dim and not bold.
                     (_, Flags::DIM) | (false, Flags::DIM_BOLD) => {
-                        self.color(ansi.to_dim() as usize, renderable_content)
+                        self.color(ansi.to_dim() as usize, term_colors)
                     }
                     // None of the above, keep original color..
-                    _ => self.color(*ansi as usize, renderable_content),
+                    _ => self.color(*ansi as usize, term_colors),
                 }
             }
             AnsiColor::Spec(rgb) => {
@@ -483,19 +523,15 @@ impl Renderer {
                     _ => *index as usize,
                 };
 
-                self.color(index, renderable_content)
+                self.color(index, term_colors)
             }
         }
     }
 
     #[inline]
-    fn compute_bg_color(
-        &self,
-        square: &Square,
-        renderable_content: &RenderableContent,
-    ) -> ColorArray {
+    fn compute_bg_color(&self, square: &Square, term_colors: &TermColors) -> ColorArray {
         match square.bg {
-            AnsiColor::Named(ansi) => self.color(ansi as usize, renderable_content),
+            AnsiColor::Named(ansi) => self.color(ansi as usize, term_colors),
             AnsiColor::Spec(rgb) => match square.flags & Flags::DIM {
                 Flags::DIM => (&(rgb * DIM_FACTOR)).into(),
                 _ => (&rgb).into(),
@@ -514,7 +550,7 @@ impl Renderer {
                     _ => idx as usize,
                 };
 
-                self.color(idx, renderable_content)
+                self.color(idx, term_colors)
             }
         }
     }
@@ -525,7 +561,7 @@ impl Renderer {
         square: &Square,
         cursor: &Cursor,
         is_active: bool,
-        renderable_content: &RenderableContent,
+        term_colors: &TermColors,
     ) -> (FragmentStyle, char) {
         let font_attrs = match (
             square.flags.contains(Flags::ITALIC),
@@ -538,8 +574,8 @@ impl Renderer {
             _ => (Stretch::NORMAL, Weight::NORMAL, Style::Normal),
         };
 
-        let mut color = self.compute_color(&square.fg, square.flags, renderable_content);
-        let mut background_color = self.compute_bg_color(square, renderable_content);
+        let mut color = self.compute_color(&square.fg, square.flags, term_colors);
+        let mut background_color = self.compute_bg_color(square, term_colors);
         // If IME is enabled we get the current content to cursor
         let content = if cursor.is_ime_enabled {
             cursor.content
@@ -586,14 +622,12 @@ impl Renderer {
         };
 
         let cursor_color = if !self.is_vi_mode_enabled {
-            renderable_content.colors[NamedColor::Cursor]
-                .unwrap_or(self.named_colors.cursor)
+            term_colors[NamedColor::Cursor].unwrap_or(self.named_colors.cursor)
         } else {
             self.named_colors.vi_cursor
         };
 
-        let (decoration, decoration_color) =
-            self.compute_decoration(square, renderable_content);
+        let (decoration, decoration_color) = self.compute_decoration(square, term_colors);
         style.decoration = decoration;
         style.decoration_color = decoration_color;
 
@@ -632,12 +666,8 @@ impl Renderer {
 
     // Get the RGB value for a color index.
     #[inline]
-    pub fn color(
-        &self,
-        color: usize,
-        renderable_content: &RenderableContent,
-    ) -> ColorArray {
-        renderable_content.colors[color].unwrap_or(self.colors[color])
+    pub fn color(&self, color: usize, term_colors: &TermColors) -> ColorArray {
+        term_colors[color].unwrap_or(self.colors[color])
     }
 
     #[inline]
@@ -707,7 +737,7 @@ impl Renderer {
     }
 
     #[inline]
-    pub fn update(
+    pub fn run(
         &mut self,
         sugarloaf: &mut Sugarloaf,
         context_manager: &mut ContextManager<EventProxy>,
@@ -722,81 +752,197 @@ impl Renderer {
             self.search.rich_text_id = Some(search_rich_text);
         }
 
-        let content = sugarloaf.content();
+        let mut graphic_queues: Option<Vec<UpdateQueues>> = None;
 
         let grid = context_manager.current_grid_mut();
         let active_index = grid.current;
+        let mut has_active_changed = false;
+        if self.last_active != active_index {
+            has_active_changed = true;
+
+            self.last_active = active_index;
+        }
 
         for (index, grid_context) in grid.contexts_mut().iter_mut().enumerate() {
             let is_active = active_index == index;
             let context = grid_context.context_mut();
+
+            let mut has_ime = false;
+            if let Some(preedit) = context.ime.preedit() {
+                if let Some(content) = preedit.text.chars().next() {
+                    context.renderable_content.cursor.content = content;
+                    context.renderable_content.cursor.is_ime_enabled = true;
+                    has_ime = true;
+                }
+            }
+
+            if !has_ime {
+                context.renderable_content.cursor.is_ime_enabled = false;
+                context.renderable_content.cursor.content =
+                    context.renderable_content.cursor.content_ref;
+            }
+
+            // let duration = start.elapsed();
+            // println!("Time elapsed in antes is: {:?}", duration);
+            // let renderable_content = context.renderable_content();
+            let force_full_damage = has_active_changed
+                || context.renderable_content.has_pending_updates
+                || is_active
+                    && (context.renderable_content.selection_range.is_some()
+                        || hints.is_some());
+
+            let mut specific_lines = None;
+            let (colors, display_offset, blinking_cursor, visible_rows) = {
+                let mut terminal = context.terminal.lock();
+                let result = (
+                    terminal.colors,
+                    terminal.display_offset(),
+                    terminal.blinking_cursor,
+                    terminal.visible_rows(),
+                );
+
+                context.renderable_content.cursor.state = terminal.cursor();
+
+                if let Some(queues_to_add) = terminal.graphics_take_queues() {
+                    if let Some(ref mut queues) = graphic_queues {
+                        queues.push(queues_to_add);
+                    } else {
+                        graphic_queues = Some(vec![queues_to_add]);
+                    }
+                }
+
+                if !force_full_damage && !terminal.is_fully_damaged() {
+                    if let TermDamage::Partial(lines) = terminal.damage() {
+                        let mut own_lines =
+                            std::collections::HashSet::with_capacity(result.3.len());
+                        for line in lines {
+                            own_lines.insert(line.line);
+                        }
+                        specific_lines = Some(own_lines);
+                    };
+                }
+                terminal.reset_damage();
+                result
+            };
+
+            // If the last line is bigger than the actual visible rows, then some resize
+            // has happened. In this case, request full draw.
+            if let Some(ref lines) = specific_lines {
+                if let Some(max_value) = lines.iter().max() {
+                    if max_value > &(visible_rows.len() - 1) {
+                        specific_lines = None;
+                    }
+                }
+            }
+
+            // let duration = start.elapsed();
+            // println!("Time elapsed in antes-antes is: {:?}", duration);
             let rich_text_id = context.rich_text_id;
-            let renderable_content = context.renderable_content();
-            let mut is_cursor_visible = renderable_content.is_cursor_visible
-                && renderable_content.cursor.state.is_visible();
-            if !is_active && renderable_content.cursor.state.is_visible() {
+
+            let mut is_cursor_visible =
+                context.renderable_content.cursor.state.is_visible();
+            context.renderable_content.has_blinking_enabled = blinking_cursor;
+            if blinking_cursor {
+                let has_selection = context.renderable_content.selection_range.is_some();
+                if !has_selection {
+                    let mut should_blink = true;
+                    if let Some(last_typing_time) = context.renderable_content.last_typing
+                    {
+                        if last_typing_time.elapsed() < std::time::Duration::from_secs(1)
+                        {
+                            should_blink = false;
+                        }
+                    }
+
+                    if should_blink {
+                        context.renderable_content.is_blinking_cursor_visible =
+                            !context.renderable_content.is_blinking_cursor_visible;
+                        if let Some(ref mut lines) = specific_lines {
+                            lines.insert(
+                                context.renderable_content.cursor.state.pos.row.0
+                                    as usize,
+                            );
+                        }
+                    } else {
+                        context.renderable_content.is_blinking_cursor_visible = true;
+                    }
+                }
+
+                is_cursor_visible = context.renderable_content.is_blinking_cursor_visible;
+            }
+
+            if !is_active && context.renderable_content.cursor.state.is_visible() {
                 is_cursor_visible = true;
             }
 
-            let display_offset = renderable_content.display_offset;
-            let strategy = if is_active && hints.is_some() {
-                &RenderableContentStrategy::Full
-            } else {
-                &renderable_content.strategy
-            };
-
-            match strategy {
-                RenderableContentStrategy::Full => {
+            context.renderable_content.has_pending_updates = false;
+            let content = sugarloaf.content();
+            match specific_lines {
+                None => {
                     content.sel(rich_text_id);
                     content.clear();
-                    for (i, row) in renderable_content.inner.iter().enumerate() {
+                    for (i, row) in visible_rows.iter().enumerate() {
                         let has_cursor = is_cursor_visible
-                            && renderable_content.cursor.state.pos.row == i;
+                            && context.renderable_content.cursor.state.pos.row == i;
                         self.create_line(
                             content,
                             row,
                             has_cursor,
                             None,
-                            Line((i as i32) - display_offset),
-                            renderable_content,
+                            Line((i as i32) - display_offset as i32),
+                            &context.renderable_content,
                             hints,
                             focused_match,
+                            &colors,
                             is_active,
                         );
                     }
                     content.build();
+                    // let duration = start.elapsed();
+                    // println!("Time elapsed in -renderer.TermDamage::Full is: {:?}", duration);
                 }
-                RenderableContentStrategy::Lines(lines) => {
+                Some(lines) => {
                     content.sel(rich_text_id);
                     for line in lines {
-                        let line = *line;
-                        if let Some(line_data) = renderable_content.inner.get(line) {
-                            let has_cursor = is_cursor_visible
-                                && renderable_content.cursor.state.pos.row == line;
-                            content.clear_line(line);
+                        let has_cursor = is_cursor_visible
+                            && context.renderable_content.cursor.state.pos.row == line;
+                        content.clear_line(line);
+                        if let Some(visible_row) = visible_rows.get(line) {
                             self.create_line(
                                 content,
-                                line_data,
+                                visible_row,
                                 has_cursor,
                                 Some(line),
-                                Line((line as i32) - display_offset),
-                                renderable_content,
+                                Line(line as i32),
+                                &context.renderable_content,
                                 hints,
                                 focused_match,
+                                &colors,
                                 is_active,
                             );
                         }
                     }
                 }
-                RenderableContentStrategy::Noop => {}
+            };
+        }
+
+        if let Some(op) = graphic_queues.take() {
+            for queues in op {
+                for graphic_data in queues.pending {
+                    sugarloaf.graphics.insert(graphic_data);
+                }
+
+                for graphic_data in queues.remove_queue {
+                    sugarloaf.graphics.remove(&graphic_data);
+                }
             }
         }
 
-        self.update_search_rich_text(content);
+        self.update_search_rich_text(sugarloaf.content());
 
         let window_size = sugarloaf.window_size();
         let scale_factor = sugarloaf.scale_factor();
-        let mut objects = Vec::with_capacity(30);
+        let mut objects = Vec::with_capacity(15);
         self.navigation.build_objects(
             sugarloaf,
             (window_size.width, window_size.height, scale_factor),
@@ -820,10 +966,11 @@ impl Renderer {
             self.search.rich_text_id = None;
         }
 
-        for rte in context_manager.grid_objects() {
-            objects.push(rte);
-        }
-
+        context_manager.extend_with_grid_objects(&mut objects);
         sugarloaf.set_objects(objects);
+
+        sugarloaf.render();
+        // let duration = start.elapsed();
+        // println!("Time elapsed in -renderer.update() is: {:?}", duration);
     }
 }
