@@ -4,23 +4,20 @@ mod image_cache;
 pub mod text;
 
 use crate::components::core::orthographic_projection;
+use crate::components::rich_text::compositor::{BatchOperation, LineCache};
 use crate::components::rich_text::image_cache::{GlyphCache, ImageCache};
 use crate::context::Context;
 use crate::font::FontLibrary;
-use crate::layout::{RichTextLayout, SugarDimensions};
+use crate::layout::{BuilderStateUpdate, RichTextLayout, SugarDimensions};
 use crate::sugarloaf::graphics::GraphicRenderRequest;
 use crate::Graphics;
-use compositor::{Compositor, DisplayList, Rect, Vertex};
+use crate::RichTextLinesRange;
+use compositor::{Compositor, Rect, Vertex};
 use std::collections::HashSet;
 use std::{borrow::Cow, mem};
 use text::{Glyph, TextRunStyle};
 use wgpu::util::DeviceExt;
 
-// Note: currently it's using Indexed drawing instead of Instance drawing could be worth to
-// evaluate if would make sense move to instance drawing instead
-// https://math.hws.edu/graphicsbook/c9/s2.html
-// https://docs.rs/wgpu/latest/wgpu/enum.VertexStepMode.html
-
 pub const BLEND: Option<wgpu::BlendState> = Some(wgpu::BlendState {
     color: wgpu::BlendComponent {
         src_factor: wgpu::BlendFactor::SrcAlpha,
@@ -41,22 +38,20 @@ pub struct RichTextBrush {
     layout_bind_group_layout: wgpu::BindGroupLayout,
     transform: wgpu::Buffer,
     pipeline: wgpu::RenderPipeline,
-    index_buffer: wgpu::Buffer,
-    index_buffer_size: u64,
     current_transform: [f32; 16],
     comp: Compositor,
-    dlist: DisplayList,
+    vertices: Vec<Vertex>,
     supported_vertex_buffer: usize,
     textures_version: usize,
     images: ImageCache,
     glyphs: GlyphCache,
+    line_cache: LineCache,
 }
 
 impl RichTextBrush {
     pub fn new(context: &Context) -> Self {
         let device = &context.device;
-        let dlist = DisplayList::new();
-        let supported_vertex_buffer = 2_000;
+        let supported_vertex_buffer = 500;
 
         let current_transform =
             orthographic_projection(context.size.width, context.size.height);
@@ -216,32 +211,22 @@ impl RichTextBrush {
         });
 
         let vertex_buffer = device.create_buffer(&wgpu::BufferDescriptor {
-            label: Some("rich_text::Instances Buffer"),
+            label: Some("rich_text::Vertices Buffer"),
             size: mem::size_of::<Vertex>() as u64 * supported_vertex_buffer as u64,
             usage: wgpu::BufferUsages::VERTEX | wgpu::BufferUsages::COPY_DST,
             mapped_at_creation: false,
         });
 
-        let index_buffer_size: &[u32] = bytemuck::cast_slice(&dlist.indices);
-        let index_buffer_size = index_buffer_size.len() as u64;
-        let index_buffer = device.create_buffer(&wgpu::BufferDescriptor {
-            label: Some("rich_text::Indices Buffer"),
-            size: index_buffer_size,
-            usage: wgpu::BufferUsages::INDEX | wgpu::BufferUsages::COPY_DST,
-            mapped_at_creation: false,
-        });
-
         RichTextBrush {
+            line_cache: LineCache::new(),
             layout_bind_group,
             layout_bind_group_layout,
             constant_bind_group,
-            index_buffer_size,
-            index_buffer,
             comp: Compositor::new(),
             images,
             textures_version: 0,
             glyphs: GlyphCache::new(),
-            dlist,
+            vertices: vec![],
             transform,
             pipeline,
             vertex_buffer,
@@ -258,34 +243,50 @@ impl RichTextBrush {
         graphics: &mut Graphics,
     ) {
         if state.rich_texts.is_empty() {
-            self.dlist.clear();
+            self.vertices.clear();
             return;
         }
 
         self.comp.begin();
         let library = state.content.font_library();
+
         for rich_text in &state.rich_texts {
             if let Some(rt) = state.content.get_state(&rich_text.id) {
+                // Check if this specific rich text needs cache invalidation
+                match &rt.last_update {
+                    BuilderStateUpdate::Full => {
+                        self.line_cache.clear_text_cache(rich_text.id);
+                    }
+                    BuilderStateUpdate::Partial(lines) => {
+                        for line in lines {
+                            self.line_cache.clear_cache(rich_text.id, line);
+                        }
+                    }
+                    BuilderStateUpdate::Noop => {
+                        // Do nothing
+                    }
+                };
+
                 let position = (
                     rich_text.position[0] * state.style.scale_factor,
                     rich_text.position[1] * state.style.scale_factor,
                 );
 
-                draw_layout(
-                    &mut self.comp,
-                    (&mut self.images, &mut self.glyphs),
+                self.draw_layout(
+                    rich_text.id, // Pass the rich text ID for caching
                     &rt.lines,
-                    position,
+                    &rich_text.lines,
+                    Some(position),
                     library,
-                    &rt.layout,
+                    Some(&rt.layout),
                     graphics,
                 );
             }
         }
 
-        self.dlist.clear();
+        self.vertices.clear();
         self.images.process_atlases(context);
-        self.comp.finish(&mut self.dlist);
+        self.comp.finish(&mut self.vertices);
     }
 
     #[inline]
@@ -293,17 +294,283 @@ impl RichTextBrush {
         &mut self,
         font_library: &FontLibrary,
         render_data: &crate::layout::BuilderLine,
+        graphics: &mut Graphics,
     ) -> Option<SugarDimensions> {
+        // Every dimension request will lead to clear lines cache
+        self.line_cache.clear_all();
         self.comp.begin();
 
-        let dimension = fetch_dimensions(
-            &mut self.comp,
-            (&mut self.images, &mut self.glyphs),
-            render_data,
-            font_library,
-        );
-        if dimension.height > 0. && dimension.width > 0. {
-            Some(dimension)
+        let lines = vec![render_data.clone()];
+        self.draw_layout(0, &lines, &None, None, font_library, None, graphics)
+    }
+
+    fn extract_font_metrics(
+        lines: &[crate::layout::BuilderLine],
+    ) -> Option<(f32, f32, f32, usize, f32)> {
+        // Extract the first run from a line that has at least one run
+        lines
+            .iter()
+            .filter(|line| !line.render_data.runs.is_empty())
+            .map(|line| &line.render_data.runs[0])
+            .next()
+            .map(|run| {
+                (
+                    run.ascent.round(),
+                    run.descent.round(),
+                    (run.leading).round() * 2.0,
+                    run.span.font_id,
+                    run.size,
+                )
+            })
+    }
+
+    #[inline]
+    #[allow(clippy::too_many_arguments)]
+    fn draw_layout(
+        &mut self,
+        rich_text_id: usize,
+        lines: &Vec<crate::layout::BuilderLine>,
+        selected_lines: &Option<RichTextLinesRange>,
+        pos: Option<(f32, f32)>,
+        font_library: &FontLibrary,
+        rte_layout: Option<&RichTextLayout>,
+        graphics: &mut Graphics,
+    ) -> Option<SugarDimensions> {
+        if lines.is_empty() {
+            return None;
+        }
+
+        // let start = std::time::Instant::now();
+        let comp = &mut self.comp;
+        let caches = (&mut self.images, &mut self.glyphs);
+        let (image_cache, glyphs_cache) = caches;
+        let font_coords: &[i16] = &[0, 0, 0, 0];
+        let depth = 0.0;
+
+        // Determine if we're calculating dimensions only or drawing layout
+        let is_dimensions_only = pos.is_none() || rte_layout.is_none();
+
+        // For dimensions mode, we only process the first line
+        let lines_to_process = if is_dimensions_only {
+            std::slice::from_ref(&lines[0])
+        } else {
+            lines.as_slice()
+        };
+
+        // Get initial position
+        let (x, y) = pos.unwrap_or((0.0, 0.0));
+
+        // Set up caches based on mode
+        let mut glyphs = Vec::new();
+        let mut last_rendered_graphic = HashSet::new();
+        let mut line_y = y;
+        let mut dimensions = SugarDimensions::default();
+
+        let font_metrics = Self::extract_font_metrics(lines_to_process);
+        if let Some((
+            ascent,
+            descent,
+            leading,
+            current_font_from_valid_run,
+            current_font_size_from_valid_run,
+        )) = font_metrics
+        {
+            // Initialize from first run if available
+            let mut current_font = current_font_from_valid_run;
+            let mut current_font_size = current_font_size_from_valid_run;
+
+            let mut session = glyphs_cache.session(
+                image_cache,
+                current_font,
+                font_library,
+                font_coords,
+                current_font_size,
+            );
+
+            // Calculate line height with modifier if available
+            let line_height_without_mod = ascent + descent + leading;
+            let line_height_mod = rte_layout.map_or(1.0, |layout| layout.line_height);
+            let line_height = line_height_without_mod * line_height_mod;
+
+            let skip_count = selected_lines.map_or(0, |range| range.start);
+            let take_count = selected_lines
+                .map_or(lines_to_process.len(), |range| range.end - range.start);
+
+            for (line_idx, line) in lines_to_process
+                .iter()
+                .enumerate()
+                .skip(skip_count)
+                .take(take_count)
+            {
+                if line.render_data.runs.is_empty() {
+                    continue;
+                }
+
+                // Check if we can use the cache for this line
+                if !is_dimensions_only
+                    && self.line_cache.has_cache(rich_text_id, line_idx)
+                    && self
+                        .line_cache
+                        .apply_cache(rich_text_id, line_idx, comp, graphics)
+                {
+                    // Cache was applied successfully, skip to next line
+                    line_y += line_height;
+                    continue;
+                }
+
+                let mut px = x;
+
+                // Calculate baseline differently based on mode
+                let baseline = if is_dimensions_only {
+                    ascent + y
+                } else {
+                    line_y + ascent
+                };
+
+                // Different line_y calculation based on mode
+                line_y = baseline + descent;
+
+                // Calculate padding
+                let padding_y = if line_height_mod > 1.0 {
+                    (line_height - line_height_without_mod) / 2.0
+                } else {
+                    0.0
+                };
+
+                let py = line_y;
+                let mut line_operations = Vec::new();
+
+                for run in &line.render_data.runs {
+                    glyphs.clear();
+                    let font = run.span.font_id;
+                    let char_width = run.span.width;
+
+                    let run_x = px;
+                    for glyph in &run.glyphs {
+                        let x = px;
+                        let y = py + padding_y;
+
+                        // Different advance calculation based on mode
+                        if is_dimensions_only {
+                            px += glyph.simple_data().1 * char_width;
+                        } else {
+                            px += rte_layout.unwrap().dimensions.width * char_width;
+                        }
+
+                        glyphs.push(Glyph {
+                            id: glyph.simple_data().0,
+                            x,
+                            y,
+                        });
+                    }
+
+                    // Create style with appropriate defaults
+                    let style = TextRunStyle {
+                        font_coords,
+                        font_size: run.size,
+                        color: run.span.color,
+                        cursor: run.span.cursor,
+                        drawable_char: run.span.drawable_char,
+                        background_color: run.span.background_color,
+                        baseline: py,
+                        topline: py - ascent,
+                        padding_y,
+                        line_height,
+                        line_height_without_mod,
+                        advance: px - run_x,
+                        decoration: run.span.decoration,
+                        decoration_color: run.span.decoration_color,
+                    };
+
+                    // Update dimensions if in dimensions mode
+                    if is_dimensions_only && style.advance > 0.0 && line_height > 0.0 {
+                        dimensions.width = style.advance.round();
+                        dimensions.height = line_height.round();
+                    }
+
+                    // Update font session if needed
+                    if font != current_font || style.font_size != current_font_size {
+                        current_font = font;
+                        current_font_size = style.font_size;
+
+                        session = glyphs_cache.session(
+                            image_cache,
+                            current_font,
+                            font_library,
+                            font_coords,
+                            style.font_size,
+                        );
+                    }
+
+                    // Handle graphics if in layout mode
+                    if !is_dimensions_only {
+                        if let Some(graphic) = run.span.media {
+                            if !last_rendered_graphic.contains(&graphic.id) {
+                                let offset_x = graphic.offset_x as f32;
+                                let offset_y = graphic.offset_y as f32;
+
+                                let graphic_render_request = GraphicRenderRequest {
+                                    id: graphic.id,
+                                    pos_x: run_x - offset_x,
+                                    pos_y: style.topline - offset_y,
+                                    width: None,
+                                    height: None,
+                                };
+
+                                graphics.top_layer.push(graphic_render_request);
+                                line_operations.push(BatchOperation::GraphicRequest(
+                                    graphic_render_request,
+                                ));
+
+                                last_rendered_graphic.insert(graphic.id);
+                            }
+                        }
+                    }
+
+                    // Use a Vec to collect operations if caching
+                    let mut run_operations = Vec::new();
+                    let cache_ops = if !is_dimensions_only {
+                        Some(&mut run_operations)
+                    } else {
+                        None
+                    };
+
+                    // Draw the run with caching if needed
+                    comp.draw_run(
+                        &mut session,
+                        Rect::new(run_x, py, style.advance, 1.),
+                        depth,
+                        &style,
+                        &glyphs,
+                        cache_ops,
+                    );
+
+                    // Add run operations to line operations
+                    if !is_dimensions_only {
+                        line_operations.extend(run_operations);
+                    }
+                }
+
+                // Store line in cache if we're not in dimensions mode
+                if !is_dimensions_only {
+                    self.line_cache
+                        .store(rich_text_id, line_idx, line_operations);
+                }
+
+                // Update line_y for line height modifier
+                if !is_dimensions_only && line_height_mod > 1.0 {
+                    line_y += line_height - line_height_without_mod;
+                }
+            }
+        }
+
+        // Return dimensions if in dimensions mode
+        if is_dimensions_only {
+            if dimensions.height > 0.0 && dimensions.width > 0.0 {
+                Some(dimensions)
+            } else {
+                None
+            }
         } else {
             None
         }
@@ -320,9 +587,8 @@ impl RichTextBrush {
         ctx: &mut Context,
         rpass: &mut wgpu::RenderPass<'pass>,
     ) {
-        // let start = std::time::Instant::now();
         // There's nothing to render
-        if self.dlist.vertices.is_empty() {
+        if self.vertices.is_empty() {
             return;
         }
 
@@ -336,12 +602,12 @@ impl RichTextBrush {
             self.current_transform = transform;
         }
 
-        if self.dlist.vertices.len() > self.supported_vertex_buffer {
+        if self.vertices.len() > self.supported_vertex_buffer {
             self.vertex_buffer.destroy();
 
-            self.supported_vertex_buffer = self.dlist.vertices.len();
+            self.supported_vertex_buffer = self.vertices.len();
             self.vertex_buffer = ctx.device.create_buffer(&wgpu::BufferDescriptor {
-                label: Some("sugarloaf::rich_text::Pipeline instances"),
+                label: Some("sugarloaf::rich_text::Pipeline vertices"),
                 size: mem::size_of::<Vertex>() as u64
                     * self.supported_vertex_buffer as u64,
                 usage: wgpu::BufferUsages::VERTEX | wgpu::BufferUsages::COPY_DST,
@@ -349,34 +615,11 @@ impl RichTextBrush {
             });
         }
 
-        let vertices_bytes: &[u8] = bytemuck::cast_slice(&self.dlist.vertices);
+        let vertices_bytes: &[u8] = bytemuck::cast_slice(&self.vertices);
         if !vertices_bytes.is_empty() {
             queue.write_buffer(&self.vertex_buffer, 0, vertices_bytes);
         }
 
-        let indices_raw: &[u8] = bytemuck::cast_slice(&self.dlist.indices);
-        let indices_raw_size = indices_raw.len() as u64;
-
-        if self.index_buffer_size >= indices_raw_size {
-            queue.write_buffer(&self.index_buffer, 0, indices_raw);
-        } else {
-            self.index_buffer.destroy();
-
-            let size = next_copy_buffer_size(indices_raw_size);
-            let buffer = ctx.device.create_buffer(&wgpu::BufferDescriptor {
-                label: Some("rich_text::Indices"),
-                size,
-                usage: wgpu::BufferUsages::INDEX | wgpu::BufferUsages::COPY_DST,
-                mapped_at_creation: true,
-            });
-            buffer.slice(..).get_mapped_range_mut()[..indices_raw.len()]
-                .copy_from_slice(indices_raw);
-            buffer.unmap();
-
-            self.index_buffer = buffer;
-            self.index_buffer_size = size;
-        }
-
         if self.textures_version != self.images.entries.len() {
             self.textures_version = self.images.entries.len();
             self.layout_bind_group =
@@ -396,262 +639,9 @@ impl RichTextBrush {
         rpass.set_bind_group(0, &self.constant_bind_group, &[]);
         rpass.set_bind_group(1, &self.layout_bind_group, &[]);
         rpass.set_vertex_buffer(0, self.vertex_buffer.slice(..));
-        rpass.set_index_buffer(self.index_buffer.slice(..), wgpu::IndexFormat::Uint32);
-        rpass.draw_indexed(0..(self.dlist.indices.len() as u32), 0, 0..1);
 
-        // let duration = start.elapsed();
-        // println!(" - rich_text::render() is: {:?}", duration);
+        // Use draw instead of draw_indexed
+        let vertex_count = self.vertices.len() as u32;
+        rpass.draw(0..vertex_count, 0..1);
     }
 }
-
-#[inline]
-fn draw_layout(
-    comp: &mut compositor::Compositor,
-    caches: (&mut ImageCache, &mut GlyphCache),
-    lines: &Vec<crate::layout::BuilderLine>,
-    pos: (f32, f32),
-    font_library: &FontLibrary,
-    rte_layout: &RichTextLayout,
-    graphics: &mut Graphics,
-) {
-    let rect = &rte_layout.dimensions;
-    let line_height_mod = &rte_layout.line_height;
-    // let start = std::time::Instant::now();
-    let (x, y) = pos;
-    let (image_cache, glyphs_cache) = caches;
-    let depth = 0.0;
-    let mut glyphs = Vec::new();
-    let mut current_font = 0;
-    let mut current_font_size = 0.0;
-    let font_coords: &[i16] = &[0, 0, 0, 0];
-    if let Some(line) = lines.first() {
-        if let Some(first_run) = line.render_data.runs.first() {
-            current_font = first_run.span.font_id;
-            current_font_size = first_run.size;
-        }
-    }
-
-    glyphs_cache.set_max_height(rect.height as u16);
-
-    let mut session = glyphs_cache.session(
-        image_cache,
-        current_font,
-        font_library,
-        font_coords,
-        current_font_size,
-    );
-
-    let mut last_rendered_graphic = HashSet::new();
-    let mut line_y = 0. + y;
-    for line in lines {
-        if line.render_data.runs.is_empty() {
-            continue;
-        }
-
-        let first_run = &line.render_data.runs[0];
-        let ascent = first_run.ascent.round();
-        let descent = first_run.descent.round();
-        let leading = (first_run.leading).round() * 2.;
-        let mut px = x + 0.0;
-        let baseline = line_y + ascent;
-        line_y = baseline + descent;
-        let line_height_without_mod = ascent + descent + leading;
-        let line_height = line_height_without_mod * line_height_mod;
-
-        // If line height is 10. and and modifier is 1.5, then 5
-        let padding_y = if line_height_mod > &1. {
-            (line_height - line_height_without_mod) / 2.
-        } else {
-            0.
-        };
-
-        let py = line_y;
-        for run in &line.render_data.runs {
-            glyphs.clear();
-            let font = run.span.font_id;
-            let char_width = run.span.width;
-
-            let run_x = px;
-            for glyph in &run.glyphs {
-                let x = px;
-                let y = py + padding_y;
-                px += rect.width * char_width;
-                glyphs.push(Glyph {
-                    id: glyph.simple_data().0,
-                    x,
-                    y,
-                });
-            }
-            let style = TextRunStyle {
-                font_coords,
-                font_size: run.size,
-                color: run.span.color,
-                cursor: run.span.cursor,
-                background_color: run.span.background_color,
-                baseline: py,
-                topline: py - ascent,
-                padding_y,
-                line_height,
-                line_height_without_mod,
-                advance: px - run_x,
-                decoration: run.span.decoration,
-                decoration_color: run.span.decoration_color,
-            };
-
-            if font != current_font || style.font_size != current_font_size {
-                current_font = font;
-                current_font_size = style.font_size;
-
-                session = glyphs_cache.session(
-                    image_cache,
-                    current_font,
-                    font_library,
-                    font_coords,
-                    style.font_size,
-                );
-            }
-
-            if let Some(graphic) = run.span.media {
-                if !last_rendered_graphic.contains(&graphic.id) {
-                    let offset_x = graphic.offset_x as f32;
-                    let offset_y = graphic.offset_y as f32;
-
-                    graphics.top_layer.push(GraphicRenderRequest {
-                        id: graphic.id,
-                        pos_x: run_x - offset_x,
-                        pos_y: style.topline - offset_y,
-                        width: None,
-                        height: None,
-                    });
-                    last_rendered_graphic.insert(graphic.id);
-                }
-            }
-
-            comp.draw_run(
-                &mut session,
-                Rect::new(run_x, py, style.advance, 1.),
-                depth,
-                &style,
-                &glyphs,
-            );
-        }
-
-        if line_height_mod > &1.0 {
-            line_y += line_height - line_height_without_mod;
-        }
-    }
-
-    // let duration = start.elapsed();
-    // println!(" - draw_layout() is: {:?}\n", duration);
-}
-
-#[inline]
-fn fetch_dimensions(
-    comp: &mut compositor::Compositor,
-    caches: (&mut ImageCache, &mut GlyphCache),
-    line: &crate::layout::BuilderLine,
-    font_library: &FontLibrary,
-) -> SugarDimensions {
-    let x = 0.;
-    let y = 0.;
-
-    let (image_cache, glyphs_cache) = caches;
-    let mut current_font = 0;
-    let mut current_font_size = 0.0;
-    let font_coords: &[i16] = &[0, 0, 0, 0];
-    if let Some(first_run) = line.render_data.runs.first() {
-        current_font = first_run.span.font_id;
-        current_font_size = first_run.size;
-    }
-
-    let mut session = glyphs_cache.session(
-        image_cache,
-        current_font,
-        font_library,
-        font_coords,
-        current_font_size,
-    );
-
-    let mut glyphs = Vec::with_capacity(3);
-    let mut dimension = SugarDimensions::default();
-    let first_run = &line.render_data.runs[0];
-    let ascent = first_run.ascent.round();
-    let descent = first_run.descent.round();
-    let leading = (first_run.leading).round() * 2.;
-    // let mut px = x + line.offset();
-    let mut px = x + 0.0;
-    let py = ascent + y;
-    let line_height = ascent + descent + leading;
-    for run in &line.render_data.runs {
-        let char_width = run.span.width;
-
-        let font = run.span.font_id;
-        let run_x = px;
-        glyphs.clear();
-        for glyph in &run.glyphs {
-            // let x = px + glyph.x;
-            // let y = py - glyph.y;
-            let x = px;
-            let y = py;
-            // px += glyph.advance
-            px += glyph.simple_data().1 * char_width;
-            glyphs.push(Glyph {
-                id: glyph.simple_data().0,
-                x,
-                y,
-            });
-        }
-        let color = run.span.color;
-
-        let style = TextRunStyle {
-            font_coords,
-            font_size: run.size,
-            color,
-            cursor: run.span.cursor,
-            background_color: None,
-            baseline: py,
-            padding_y: 0.,
-            topline: py - ascent,
-            line_height,
-            line_height_without_mod: line_height,
-            advance: px - run_x,
-            decoration: None,
-            decoration_color: None,
-        };
-
-        if style.advance > 0. && line_height > 0. {
-            dimension.width = style.advance.round();
-            dimension.height = line_height.round();
-        }
-
-        if font != current_font || style.font_size != current_font_size {
-            current_font = font;
-            current_font_size = style.font_size;
-
-            session = glyphs_cache.session(
-                image_cache,
-                current_font,
-                font_library,
-                font_coords,
-                current_font_size,
-            );
-        }
-
-        comp.draw_run(
-            &mut session,
-            Rect::new(run_x, py, style.advance, 1.),
-            0.0,
-            &style,
-            &glyphs,
-        );
-    }
-
-    dimension
-}
-
-#[inline]
-fn next_copy_buffer_size(size: u64) -> u64 {
-    let align_mask = wgpu::COPY_BUFFER_ALIGNMENT - 1;
-    ((size.next_power_of_two() + align_mask) & !align_mask)
-        .max(wgpu::COPY_BUFFER_ALIGNMENT)
-}
